#!/usr/bin/env python

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

from collections import deque
from math import ceil
import sys

from intervention import intervention_at_time
from nslookup4 import lookup_or_die
from scanerror import ScanError
from sequence import TCPSequence
from sniffing import start_sniffing4 as start_sniffing
from synack4 import synack_retransmit_times
from tcp4 import receive_tcp
from tcptest4 import tcp_detect_ip_id_pattern, tcp_test_rate

ALPHA = 0.01
INTERVENTION = 20
SAMPLES = 100
SLEEP_SMALL = 0.1
TIMEOUT = 1.0
WINDOW = 20

def test_censor_case(sniffer, saddr, caddr, spoofed_addr, port):
    last_ip_id = None
    xs = []
    pattern = tcp_detect_ip_id_pattern(sniffer, caddr, spoofed_addr, sport=port, flags='SA')
    if pattern['type'] != 'globally':
        raise ScanError('Client has \'%s\' IP id pattern, not globally incrementing' % pattern['type'])
    retransmit_times, avg_trans, syn_frac = synack_retransmit_times(sniffer, saddr, spoofed_addr, port, window=WINDOW, repeat=INTERVENTION, verbose=True)
    sys.stderr.write('Retransmit times: %s\n' % retransmit_times)
    if not retransmit_times:
        raise ScanError('Cannot get SYNACK retransmit times from server on port %d' % port)
    if len(retransmit_times) < 2:
        raise ScanError('Server on port %d must transmit at least two SYNACK\'s' % port)
    if syn_frac <= 0.5 or avg_trans <= (len(retransmit_times) + 1.0) / 2.0:
        raise ScanError('Server on port %d responds with too few SYNACK transmissions' % port)
    rate = tcp_test_rate(sniffer, caddr, len(retransmit_times) * INTERVENTION, sport=port, flags='SA')
    if rate['avg'] <= (len(retransmit_times) + 1.0) / 2.0:
        raise ScanError('Client responds with too few RST\'s')
    retransmit_times = [int(ceil(t)) for t in retransmit_times]
    retransmit_times[0] = 0
    seq = TCPSequence(caddr, iface=sniffer.iface)
    seq_spoofed = TCPSequence(saddr, caddr)
    timeout = sniffer.timeout()
    reset_delay = retransmit_times[-1] + 1
    ts = [t + SAMPLES // 2 for t in retransmit_times]
    rstss = deque()
    for i in xrange(SAMPLES + 1):
        if i > SAMPLES // 2:
            requests, rsts = zip(*packets)
            for request in requests:
                sniffer.send(request)
            rstss.append(rsts)
        timeout.add_time(SLEEP_SMALL)
        timeout.sleep()
        x = seq.next_packet(sport=port, flags='SA')
        sniffer.send(x)
        if i > SAMPLES // 2 + reset_delay:
            rsts = rstss.popleft()
            for rst in rsts:
                sniffer.send(rst)
        if SAMPLES // 2 <= i < SAMPLES:
            # generate next packets (expensive) before next loop
            packets = [seq_spoofed.next_packet_with_rst(dport=port) for _ in xrange(INTERVENTION)]
        sniffer.work()
        timeout.add_time(TIMEOUT - SLEEP_SMALL)
        p = receive_tcp(sniffer, x, timeout)
        ip_id = None if p is None else p.id
        if i > 0:
            diff = None if ip_id is None or last_ip_id is None else ip_id - last_ip_id - 1 & 0xffff
            sys.stderr.write('%s%s: %s\n' % (ip_id, '*' * sum(i > t for t in ts), diff))
            xs.append(diff)
        last_ip_id = ip_id
        timeout.sleep()
    while rstss:
        rsts = rstss.popleft()
        for rst in rsts:
            sniffer.send(rst)
    try:
        ir = intervention_at_time(xs, ts, [0.5 * INTERVENTION, (len(ts) + 1.0) / 2.0 * INTERVENTION], 0.05, True)
    except Exception:
        raise ScanError('Client traffic is too noisy or incomplete--try another client?')
    sys.stderr.write('%s\n' % ir)
    if ir.pvalues[0] < ALPHA:
        return False
    elif ir.pvalues[0] > 1.0 - ALPHA:
        if ir.pvalues[1] < ALPHA:
            return None
        elif ir.pvalues[1] > 1.0 - ALPHA:
            return True
    raise ScanError('Client traffic is too noisy or incomplete--try another client?')

def usage(code=1):
    sys.stderr.write('Usage: %s SERVER PORT CLIENT SPOOFED_HOST\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 5:
            usage()
        shost = sys.argv[1]
        try:
            port = int(sys.argv[2])
        except ValueError:
            usage()
        chost = sys.argv[3]
        spoofed_host = sys.argv[4]
        saddr = lookup_or_die(shost)[0]
        caddr = lookup_or_die(chost)[0]
        spoofed_addr = lookup_or_die(spoofed_host)[0]
        is_server_to_client = start_sniffing(test_censor_case, saddr, caddr, spoofed_addr, port)
        if is_server_to_client is None:
            print 'No censorship'
        elif is_server_to_client:
            print 'Server to client censorship'
        else:
            print 'Client to server censorship'
    except ScanError as e:
        sys.stderr.write('Error performing scan: %s\n' % e)
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
